"""
Scaled gradients for backward pass.
This also covers gradient reversal, which is simply the case with scale=-1.
We actually extend the simple scaling by some further optional transformations like shifting.

The code is adapted from our TF implementation, see :func:`returnn.tf.util.basic.scaled_gradient`.

For some discussion on the specific implementation, see:
https://discuss.pytorch.org/t/gradient-scaling-reversal/186392

Also see other reference implementations:
https://github.com/facebookresearch/fairseq/blob/100cd91db19bb/fairseq/modules/grad_multiply.py
https://github.com/janfreyberg/pytorch-revgrad/blob/449fa763a76d/src/pytorch_revgrad/functional.py
https://github.com/tadeephuy/GradientReversal/blob/5d9857d63/gradient_reversal/functional.py
"""

from __future__ import annotations
from typing import Optional, Union
import torch


# noinspection PyMethodOverriding,PyAbstractClass,PyMissingOrEmptyDocstring
class _ScaledGradient(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x: torch.Tensor, scale: float) -> torch.Tensor:
        ctx.scale = scale
        return x

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output * ctx.scale, None


def scaled_gradient(x: torch.Tensor, scale: float) -> torch.Tensor:
    """
    :param x:
    :param scale:
    :return: just x, however, in backward pass, the gradient is scaled by the given factor
    """
    return _ScaledGradient.apply(x, scale)


# noinspection PyMethodOverriding,PyAbstractClass,PyMissingOrEmptyDocstring
class _ScaledGradientExt(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        x: torch.Tensor,
        scale: Union[float, torch.Tensor] = 1.0,
        shift: Optional[Union[float, torch.Tensor]] = None,
        scale_shift_by_sum_over_axis: Optional[int] = None,
    ):
        ctx.scale = scale
        ctx.shift = shift
        ctx.scale_shift_by_sum_over_axis = scale_shift_by_sum_over_axis
        return x

    @staticmethod
    def backward(ctx, grad):
        grad_out = grad
        if isinstance(ctx.scale, torch.Tensor) or ctx.scale != 1:
            grad_out = grad_out * ctx.scale
        if ctx.shift is not None and (isinstance(ctx.shift, torch.Tensor) or ctx.shift != 0):
            if ctx.scale_shift_by_sum_over_axis is not None:
                m = torch.sum(torch.abs(grad), dim=ctx.scale_shift_by_sum_over_axis, keepdim=True)
                grad_out = grad_out + ctx.shift * m
            else:
                grad_out = grad_out + ctx.shift
        return grad_out, None, None, None


def scaled_gradient_ext(
    x: torch.Tensor,
    *,
    scale: Union[float, torch.Tensor] = 1.0,
    shift: Optional[Union[float, torch.Tensor]] = None,
    scale_shift_by_sum_over_axis: Optional[int] = None,
):
    """
    :param x:
    :param scale: will scale gradient by this value
    :param shift: will shift gradient by this value
    :param scale_shift_by_sum_over_axis: if given, will scale and shift by the sum over the given axis
    :return: just x, but gradient in backward pass will be transformed accordingly
    """
    return _ScaledGradientExt.apply(x, scale, shift, scale_shift_by_sum_over_axis)



"""
Stepwise scheduler, e.g. for learning rate or other hyperparameters.

All these modules will accept any args/kwargs but leave them unused,
and instead uses :func:`get_run_ctx` to get the current train step from the current run context.
"""

from __future__ import annotations
from returnn.tensor import Tensor
import returnn.frontend as rf
from .piecewise_linear import PiecewiseLinear


__all__ = ["PiecewiseLinearStepwiseScheduler"]


class PiecewiseLinearStepwiseScheduler(PiecewiseLinear):
    """
    Piecewise linear scheduler based on the current global train step.

    Example::

        scheduler = PiecewiseLinearStepwiseScheduler(
            {0: 1.0, 10000: 0.1, 20000: 0.01}
        )

    This will start with 1.0, and then linearly decay to 0.1 at step 10000, and then to 0.01 at step 20000.
    """

    def __call__(self, *args, **kwargs) -> Tensor:
        return super().__call__(rf.get_run_ctx_step())



"""
Reduce
"""

from __future__ import annotations
from typing import Optional, Union, TypeVar, Sequence, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf

T = TypeVar("T")

__all__ = [
    "reduce",
    "reduce_sum",
    "reduce_max",
    "reduce_min",
    "reduce_mean",
    "reduce_logsumexp",
    "reduce_logmeanexp",
    "reduce_any",
    "reduce_all",
    "reduce_argmin",
    "reduce_argmax",
    "reduce_out",
    "RunningMean",
    "top_k",
]


def reduce(
    source: Tensor[T],
    *,
    mode: str,
    axis: Union[Dim, Sequence[Dim]],
    use_mask: bool = True,
) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param mode: "sum", "max", "min", "mean", "logsumexp", "any", "all", "argmin", "argmax"
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    # noinspection PyProtectedMember
    return source._raw_backend.reduce(source=source, mode=mode, axis=axis, use_mask=use_mask)


def reduce_sum(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="sum", axis=axis, use_mask=use_mask)


def reduce_max(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="max", axis=axis, use_mask=use_mask)


def reduce_min(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="min", axis=axis, use_mask=use_mask)


def reduce_mean(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="mean", axis=axis, use_mask=use_mask)


def reduce_logsumexp(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    Also see :func:`log_add_exp`.

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="logsumexp", axis=axis, use_mask=use_mask)


def reduce_logmeanexp(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    s = reduce_logsumexp(source, axis=axis, use_mask=use_mask)
    return s - rf.safe_log(rf.cast(rf.num_elements_of_shape(axis, use_mask=use_mask, device=s.device), s.dtype))


def reduce_any(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="any", axis=axis, use_mask=use_mask)


def reduce_all(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="all", axis=axis, use_mask=use_mask)


def reduce_argmin(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="argmin", axis=axis, use_mask=use_mask)


def reduce_argmax(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="argmax", axis=axis, use_mask=use_mask)


def reduce_out(
    source: Tensor,
    *,
    mode: str,
    num_pieces: int,
    out_dim: Optional[Dim] = None,
) -> Tensor:
    """
    Combination of :class:`SplitDimsLayer` applied to the feature dim
    and :class:`ReduceLayer` applied to the resulting feature dim.
    This can e.g. be used to do maxout.

    :param source:
    :param mode: "sum" or "max" or "mean"
    :param num_pieces: how many elements to reduce. The output dimension will be input.dim // num_pieces.
    :param out_dim:
    :return: out, with feature_dim set to new dim
    """
    assert source.feature_dim
    parts_dim = Dim(num_pieces, name="readout-parts")
    if not out_dim:
        out_dim = source.feature_dim // parts_dim
    out = rf.split_dims(source, axis=source.feature_dim, dims=(out_dim, parts_dim))
    out = reduce(out, mode=mode, axis=parts_dim)
    out.feature_dim = out_dim
    return out


class RunningMean(rf.Module):
    """
    Running mean, using exponential moving average, using the formula::

        # E.g. for some input [B,T,F], reduce to [F], when the mean vector is [F].
        new_value = reduce_mean(new_value, axis=[d for d in x.dims if d not in mean.dims])

        new_mean = alpha * new_value + (1 - alpha) * old_mean
                 = old_mean + alpha * (new_value - old_mean)  # more numerically stable

    (Like the TF :class:`AccumulateMeanLayer`.)
    (Similar is also the running mean in :class:`BatchNorm`.)
    """

    def __init__(
        self,
        in_dim: Union[Dim, Sequence[Dim]],
        *,
        alpha: float,
        dtype: Optional[str] = None,
        is_prob_distribution: Optional[bool] = None,
        update_only_in_train: bool = True,
    ):
        """
        :param in_dim: the dim of the mean vector, or the shape.
        :param alpha: factor for new_value. 0.0 means no update, 1.0 means always the new value.
            Also called momentum. E.g. 0.1 is a common value, or less, like 0.001.
        :param dtype: the dtype of the mean vector
        :param is_prob_distribution: if True, will initialize the mean vector with 1/in_dim.
        :param update_only_in_train: if True (default), will only update the mean vector in training mode.
            False means it will always update.
        """
        super().__init__()
        self.in_dim = in_dim
        self.shape = (in_dim,) if isinstance(in_dim, Dim) else in_dim
        assert all(isinstance(d, Dim) for d in self.shape)
        self.alpha = alpha
        self.is_prob_distribution = is_prob_distribution
        self.mean = rf.Parameter(self.shape, dtype=dtype, auxiliary=True)
        if is_prob_distribution:
            assert in_dim.dimension is not None
            self.mean.initial = 1.0 / in_dim.dimension
        self.update_only_in_train = update_only_in_train

    def __call__(self, x: Tensor) -> Tensor:
        """
        :param x: shape [..., F]
        :return: shape [F]
        """

        def _update_running_stats():
            assert all(d in x.dims for d in self.shape)
            x_ = rf.reduce_mean(x, axis=[d for d in x.dims if d not in self.shape])
            self.mean.assign_add(self.alpha * (x_ - self.mean))

        rf.cond(
            (not self.update_only_in_train) or rf.get_run_ctx().is_train_flag_enabled(func=RunningMean.__call__),
            _update_running_stats,
            lambda: None,
        )
        return self.mean


# noinspection PyShadowingBuiltins
def top_k(
    source: Tensor,
    *,
    axis: Union[Dim, Sequence[Dim]],
    k: Optional[Union[int, Tensor]] = None,
    k_dim: Optional[Dim] = None,
    sorted: bool = True,
) -> Tuple[Tensor, Union[Tensor, Sequence[Tensor]], Dim]:
    """
    Basically wraps tf.nn.top_k.
    Returns the top_k values and the indices.

    For an input [B,D] with axis=D, the output and indices values are shape [B,K].

    It's somewhat similar to :func:`reduce` with max and argmax.
    The axis dim is reduced and then a new dim for K is added.

    Axis can also cover multiple axes, such as [beam,classes].
    In that cases, there is not a single "indices" sub-layer,
    but sub-layers "indices0" .. "indices{N-1}"
    corresponding to each axis, in the same order.

    All other axes are treated as batch dims.

    :param source:
    :param axis: the axis to do the top_k on, which is reduced, or a sequence of axes
    :param k: the "K" in "TopK"
    :param k_dim: the new axis dim for K. if not provided, will be automatically created.
    :param sorted:
    :return: values, indices (sequence if axis is a sequence), k_dim
    """
    if k is None:
        assert k_dim, "top_k: either provide `k` or `k_dim`"
        k = k_dim.dimension or k_dim.dyn_size_ext
        assert k is not None, f"top_k: k_dim {k_dim} undefined and no k provided"
    # noinspection PyProtectedMember
    return source._raw_backend.top_k(source, axis=axis, k=k, k_dim=k_dim, sorted=sorted)



"""
Helpers to debug nan/inf values in tensors.
E.g., you get nan/inf values in the loss, and you want to know where it comes from.
There could be multiple potential sources:

- The parameters are already broken (nan/inf).
    Then some prev step caused this.
    For this, we might want to add another option which performs a check before we update params,
    so that updating params will never break them unnoticed.
- The gradients are broken (nan/inf).
    There are some PyTorch utilities to check this.
    This is currently not the focus here.
- Some part of the (forward) computation results in nan/inf.
    Currently, this is the focus here.
    We want to know where this happens.

We could run the forward pass again in different modes:

- Python tracing, and inspecting all local variables which are tensors.
    (Probably slow).
- PyTorch JIT tracing to compute the loss. This will give us the computation graph.
    We can run this computation graph again and inspect all the intermediate values,
    and then see where the nan/inf values come from.
- PyTorch profiling.

Note, one problem is non-determinism in the computation via e.g. dropout.
So the method might not be totally reliable.
Also, there might be inf/nan values which are ok, expected, and not a problem
(e.g. masking the logits for attention).
So we don't stop on the first occurrence but just report all of them.
"""

from __future__ import annotations

import sys
from typing import Optional, Union, Callable, TextIO
from io import TextIOBase
import traceback
from types import FrameType
import torch
import tree

# noinspection PyProtectedMember
from torch.utils._python_dispatch import TorchDispatchMode

# noinspection PyProtectedMember
from torch._dispatch.python import no_python_dispatcher


def debug_inf_nan(
    func: Callable[[], Optional[torch.Tensor]],
    *,
    with_grad: bool = False,
    report_every_op_call: bool = True,
    stop_reporting_after_first_inf_nan: bool = True,
    file: Optional[Union[TextIO, TextIOBase]] = None,
):
    """
    Debug the function.

    :param func: will be called like func(). if `with_grad`, we expect some loss tensor as return,
        and we will call `loss = func(); loss.backward()`.
    :param with_grad: whether to compute and debug gradients for inf/nan.
    :param report_every_op_call: whether to report every op call.
    :param stop_reporting_after_first_inf_nan: whether to stop reporting after the first inf/nan.
    :param file: where to write the output to. Default is stdout.
    """

    if file is None:
        file = sys.stdout

    # noinspection PyUnresolvedReferences,PyProtectedMember
    cur_frame: FrameType = sys._getframe()
    trace_ops = _TraceOps(
        root_frame=cur_frame,
        file=file,
        report_every_op_call=report_every_op_call,
        stop_reporting_after_first_inf_nan=stop_reporting_after_first_inf_nan,
    )

    if with_grad:
        with torch.autograd.detect_anomaly():
            with trace_ops:  # currently only for forward (but we might want to trace the backward too)
                loss = func()
            file.flush()  # the backward detect_anomaly might screw up the output otherwise
            try:
                loss.backward()
            except RuntimeError as exc:
                print(f"Caught RuntimeError in backward: {exc}", file=file)

    else:  # without grad
        with trace_ops:
            func()


# For efficiency, and to be less spammy
_TraceFuncNameBlacklist = {
    "aten::empty.memory_format",
    "aten::zeros_like",
    "aten::ones_like",
    "aten::full",
    "aten::scalar_tensor",  # when we deliberately create a scalar inf tensor
    "aten::_local_scalar_dense",
    "aten::where.self",  # when we intentionally mask with inf
    "aten::detach",
    "aten::_to_copy",
    "aten::clone",
    "aten::stack",
    "aten::view",
    "aten::_unsafe_view",
    "aten::permute",
    "aten::t",
    "aten::split_with_sizes",
    "aten::slice.Tensor",
    "aten::select.int",
    "aten::max_pool2d_with_indices",
}


class _TraceOps(TorchDispatchMode):
    def __init__(
        self,
        *,
        root_frame: FrameType,
        file: Union[TextIO, TextIOBase],
        report_every_op_call: bool = True,
        stop_reporting_after_first_inf_nan: bool = True,
    ):
        super().__init__()
        self.root_frame = root_frame
        self.file = file
        self.enabled = True
        self.report_every_op_call = report_every_op_call
        self.stop_reporting_after_first_inf_nan = stop_reporting_after_first_inf_nan

    def __torch_dispatch__(self, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if not self.enabled or func.name() in _TraceFuncNameBlacklist:
            return func(*args, **kwargs)
        if self.report_every_op_call:
            print(f"--- op {func.name()}", file=self.file)
        out = func(*args, **kwargs)
        for out_ in tree.flatten(out):
            if isinstance(out_, torch.Tensor):
                with no_python_dispatcher():
                    got_nan_inf_t = torch.stack([torch.isnan(out_).any(), torch.isinf(out_).any()]).cpu()
                    got_nan = got_nan_inf_t[0].item()
                    got_inf = got_nan_inf_t[1].item()
                    if got_nan or got_inf:
                        s = "/".join([s_ for s_, b in [("nan", got_nan), ("inf", got_inf)] if b])
                        print(f"--> {s} in {func}: {out_}", file=self.file)
                        traceback.print_list(
                            _extract_stack_up_to(skip_top_num_frames=1, root_frame=self.root_frame), file=self.file
                        )
                        if self.stop_reporting_after_first_inf_nan:
                            self.enabled = False
        return out


def _walk_stack_up_to(f: FrameType, *, root_frame: FrameType):
    while f is not None and f != root_frame:
        yield f, f.f_lineno
        f = f.f_back


def _extract_stack_up_to(*, skip_top_num_frames: int = 0, root_frame: FrameType):
    # noinspection PyUnresolvedReferences,PyProtectedMember
    frame = sys._getframe()
    skip_top_num_frames += 1  # skip this function
    for _ in range(skip_top_num_frames):
        frame = frame.f_back
    stack = traceback.StackSummary.extract(_walk_stack_up_to(frame, root_frame=root_frame))
    stack.reverse()
    return stack



#!/usr/bin/env python3

"""
Average model checkpoints.

References:
    Our :file:`tf_avg_checkpoins.py`.
    https://github.com/espnet/espnet/blob/master/utils/average_checkpoints.py
    https://github.com/facebookresearch/fairseq/blob/main/scripts/average_checkpoints.py
"""

from __future__ import annotations

import os
from typing import Optional, Any, Sequence, Dict
from collections import defaultdict
import argparse
import torch


def main():
    """main entry"""
    print(f"RETURNN {os.path.basename(__file__)} -- average PyTorch model checkpoints")
    print("PyTorch version:", torch.__version__)

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "--checkpoints", nargs="+", required=True, help="comma-separated (or multiple provided) input checkpoints"
    )
    arg_parser.add_argument("--separator", default=",", help="custom separator for --checkpoints, can also disable it")
    arg_parser.add_argument("--prefix", default="", help="add this as a prefix to the input checkpoints")
    arg_parser.add_argument("--postfix", default="", help="add this as a postfix to the input checkpoints (e.g. '.pt')")
    arg_parser.add_argument("--output_path", required=True, help="output checkpoint")
    args = arg_parser.parse_args()

    in_ckpts = []

    def _add_in_ckpt(name: str):
        in_ckpt__ = args.prefix + name + args.postfix
        if not os.path.exists(in_ckpt__):
            raise Exception(
                f"input checkpoint not found: {in_ckpt__!r}, "
                f"prefix {args.prefix!r}, postfix {args.postfix!r}, name {name!r}"
            )
        print("in ckpt:", in_ckpt__)
        in_ckpts.append(in_ckpt__)

    for in_ckpt in args.checkpoints:
        in_ckpt: str
        if args.separator:
            for in_ckpt_ in in_ckpt.split(args.separator):
                _add_in_ckpt(in_ckpt_)
        else:
            _add_in_ckpt(in_ckpt)

    print("out ckpt:", args.output_path)
    merge_checkpoints(in_ckpts=in_ckpts, out_ckpt=args.output_path)
    print("Done.")


def merge_checkpoints(in_ckpts: Sequence[str], out_ckpt: str, extra_state: Optional[Dict[str, Any]] = None):
    """
    Merge checkpoints
    """
    out_model_state: Dict[str, torch.Tensor] = {}
    out_model_state_num: Dict[str, int] = defaultdict(int)
    out_state: Dict[str, Any] = {"model": out_model_state, "merged_epochs": [], "merged_steps": []}
    for in_ckpt in in_ckpts:
        print("read ckpt:", in_ckpt)
        torch_version = tuple(int(s) for s in str(torch.__version__).split(".")[:2])
        load_kwargs = dict(map_location=torch.device("cpu"), mmap=True)
        if torch_version < (2, 1):
            load_kwargs.pop("mmap")  # mmap flag only introduced from 2.1.0 onwards
        in_state = torch.load(in_ckpt, **load_kwargs)
        assert isinstance(in_state, dict)

        assert "model" in in_state and isinstance(in_state["model"], dict)
        covered_keys = {"model"}
        for k, v in in_state["model"].items():
            k: str
            v: torch.Tensor
            if k in out_model_state:
                out_model_state[k] += v
            else:
                out_model_state[k] = v
            out_model_state_num[k] += 1

        # Take max value of the following
        for k in ["epoch", "step"]:
            covered_keys.add(k)
            if k in in_state:
                if k in out_state:
                    out_state[k] = max(out_state[k], in_state[k])
                else:
                    out_state[k] = in_state[k]

        # Extend list of following
        for k in ["merged_epochs", "merged_steps"]:
            covered_keys.add(k)
            if k in in_state:
                out_state[k].extend(in_state[k])
        out_state["merged_epochs"].append(in_state.get("epoch"))
        out_state["merged_steps"].append(in_state.get("step"))

        # Take first value of all remaining keys
        for k, v in in_state.items():
            if k in covered_keys:
                continue
            if k not in out_state:
                out_state[k] = v

    # Average
    for k in out_model_state:
        out_model_state[k] = out_model_state[k] / out_model_state_num[k]

    if extra_state:
        out_state.update(extra_state)

    print("save ckpt:", out_ckpt)
    torch.save(out_state, out_ckpt)


if __name__ == "__main__":
    main()



"""
This file is going to be imported by Debug.debug_shell() and available as interactive commands.
"""

import sys
import numpy


def find_obj_in_stack(cls, stack=None, all_threads=True):
    """
    :param type cls:
    :param types.FrameType|traceback.FrameSummary stack:
    :param bool all_threads:
    :return: obj
    """
    if all_threads:
        assert stack is None
        # noinspection PyProtectedMember,PyUnresolvedReferences
        for tid, stack in sys._current_frames().items():
            obj = find_obj_in_stack(cls=cls, stack=stack, all_threads=False)
            if obj is not None:
                return obj
        return None

    assert not all_threads
    if stack is None:
        # noinspection PyProtectedMember,PyUnresolvedReferences
        stack = sys._getframe()
        assert stack, "could not get stack"

    import inspect

    isframe = inspect.isframe

    _tb = stack
    while _tb is not None:
        if isframe(_tb):
            f = _tb
        else:
            f = _tb.tb_frame

        for obj in f.f_locals.values():
            if isinstance(obj, cls):
                return obj

        if isframe(_tb):
            _tb = _tb.f_back
        else:
            _tb = _tb.tb_next

    return None


class SimpleHdf:
    """
    Simple HDF writer.
    """

    def __init__(self, filename):
        import h5py

        self.hdf = h5py.File(filename)
        self.seq_tag_to_idx = {name: i for (i, name) in enumerate(self.hdf["seqTags"])}
        self.num_seqs = len(self.hdf["seqTags"])
        assert self.num_seqs == len(self.seq_tag_to_idx), "not unique seq tags"
        seq_lens = self.hdf["seqLengths"]
        if seq_lens.ndim == 2:
            seq_lens = seq_lens[:, 0]
        assert self.num_seqs == len(seq_lens)
        self.seq_starts = [0] + list(numpy.cumsum(seq_lens))
        total_len = self.seq_starts[-1]
        inputs_len = self.hdf["inputs"].shape[0]
        assert total_len == inputs_len, "time-dim does not match: %i vs %i" % (total_len, inputs_len)
        assert self.seq_starts[-1] == self.hdf["targets/data/classes"].shape[0]

    def get_seq_tags(self):
        """
        :rtype: list[str]
        """
        return self.hdf["seqTags"]

    def get_data(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: numpy.ndarray
        """
        seq_t0, seq_t1 = self.seq_starts[seq_idx : seq_idx + 2]
        return self.hdf["inputs"][seq_t0:seq_t1]

    def get_targets(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: numpy.ndarray
        """
        seq_t0, seq_t1 = self.seq_starts[seq_idx : seq_idx + 2]
        return self.hdf["targets/data/classes"][seq_t0:seq_t1]

    def get_data_dict(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: dict[str,numpy.ndarray]
        """
        return {"data": self.get_data(seq_idx), "classes": self.get_targets(seq_idx)}



"""
Cache, to store some data.
See :class:`Cache`.

One use case example is :func:`sinusoidal_positional_encoding` and :func:`relative_positional_encoding`.
"""

from __future__ import annotations
from typing import Optional, Union, Any, Type, Callable, Tuple, Dict, List
from weakref import ref
import tree
from returnn.util.lru_cache import lru_cache
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from returnn.frontend._backend import global_backend, get_backend_by_raw_tensor_type, Backend


class Cache:
    """
    Cache, intended for internal use of RF functions.

    One use case example is :func:`sinusoidal_positional_encoding` and :func:`relative_positional_encoding`.

    There are some specific properties we must take care of:

    - Lifetime of values: For graph-based backends, it can only stay alive for the current run ctx.
      (For eager-based backends, there is no such restriction.)
    - Size: Put some limit, use LRU logic.
    - Dims: Use only weakrefs. Some Dim should not stay alive just because of the cache.
    - Scalar dynamic Dims in eager mode, or static dims: Instead of the Dim, use the dim value for the key
      (and map the output to the Dim).
    - Tensor as keys: Use weakrefs. Also don't check by value but by identity.
    """

    def __init__(self, max_size: int):
        # Use lru_cache here, but not via a decorator,
        # as we want custom set/get logic.
        # Also, we want the lru_cache to be local to this Cache instance,
        # not shared over all instances of this class.
        self._lru_cache = lru_cache(max_size)(_lru_cache_dummy_func)

    def __repr__(self):
        return f"<{self.__class__.__module__}.{self.__class__.__qualname__} {self._lru_cache.cache_info()}>"

    def get(self, key, default=None):
        """
        :param key:
        :param default:
        :return: entry in cache or default
        """
        key_transformed = _transform_key(key)
        key_transformed_orig, value = self._lru_cache.cache_peek(key_transformed, fallback=(None, None))
        if key_transformed_orig is None:
            return default

        assert len(key_transformed_orig) == len(key_transformed)
        dim_map = {}  # orig -> new
        for key_item_orig, key_item in zip(key_transformed_orig, key_transformed):
            if isinstance(key_item_orig, DimWrapper):
                assert isinstance(key_item, DimWrapper)
                dim_orig = key_item_orig.dim_ref()
                if dim_orig is None:  # orig dim could be dead. but then it would not be used anyway
                    continue
                dim = key_item.dim_ref()
                assert isinstance(dim_orig, Dim) and isinstance(dim, Dim)
                dim_map[dim_orig] = dim

        # noinspection PyShadowingNames
        def _map_output(output):
            if isinstance(output, Dim):
                return dim_map.get(output, output)
            if isinstance(output, Tensor):
                if any(dim in dim_map for dim in output.dims):
                    out_raw = output.raw_tensor
                    for axis, dim in enumerate(output.dims):
                        if dim in dim_map:
                            output = output.copy_template_replace_dim_tag(axis=axis, new_dim_tag=dim_map[dim])
                    output.raw_tensor = out_raw
            return output

        return tree.map_structure(_map_output, value)

    def set(self, key, value):
        """
        :param key:
        :param value:
        """

        def _finalize_callback(*_args):
            self._lru_cache.cache_pop(key_transformed, fallback=None)

        key_backend = _get_backend(key)
        value_backend = _get_backend(value)
        if key_backend != value_backend:
            raise ValueError(f"key and value have different backends: {key_backend} != {value_backend}")
        key_transformed = _transform_key(key, finalize_callback=_finalize_callback)
        self._lru_cache.cache_set(key_transformed, result=(key_transformed, value))


def _lru_cache_dummy_func(*_args, **_kwargs):
    raise Exception("This should not be called.")


def _transform_key(
    key: Any, *, finalize_callback: Optional[Callable] = None, collected_dim_map: Optional[Dict[Dim, DimWrapper]] = None
) -> Tuple[Union[Type[Backend], ref[rf.RunCtx], _KeyItemType], ...]:
    backend = _get_backend(key)
    keys_flat: List[Any] = [backend]
    if not backend.executing_eagerly():
        # See comment above: If graph-mode, the cached value becomes invalid
        # when the current run ctx goes out of scope.
        keys_flat.append(ref(rf.get_run_ctx(), finalize_callback))
    if collected_dim_map is None:
        collected_dim_map = {}
    keys_flat += [
        _transform_key_item(key, finalize_callback=finalize_callback, collected_dim_map=collected_dim_map)
        for key in tree.flatten(key)
    ]
    return tuple(keys_flat)


def _transform_key_item(
    key: Any, *, finalize_callback: Optional[Callable] = None, collected_dim_map: Dict[Dim, DimWrapper]
) -> _KeyItemType:
    if isinstance(key, Tensor):
        return TensorWrapper(key, finalize_callback=finalize_callback)
    if isinstance(key, Dim):
        if key in collected_dim_map:
            return collected_dim_map[key]
        dim_wrapper = DimWrapper(key, finalize_callback=finalize_callback)
        collected_dim_map[key] = dim_wrapper
        return dim_wrapper
    if not isinstance(key, _RawTypes):
        raise TypeError(f"unexpected type {type(key)}")
    return key


def _get_backend(*args) -> Type[Backend]:
    args_flat = tree.flatten(args)
    for arg in args_flat:
        if isinstance(arg, Tensor) and arg.raw_tensor is not None:
            return get_backend_by_raw_tensor_type(type(arg.raw_tensor))
    return global_backend.__class__


class TensorWrapper:
    """
    Wraps :class:`Tensor`.
    Using weakref for the tensor, including also ``raw_tensor``.
    Equality is given if the identity is the same, for the Tensor itself and the raw_tensor.
    No value of the tensor is checked.
    """

    def __init__(self, value: Tensor, *, finalize_callback):
        self.value_ref = ref(value, finalize_callback)
        self.raw_value_ref = ref(value.raw_tensor, finalize_callback)
        self._hash = id(value)

    def __eq__(self, other):
        if isinstance(other, TensorWrapper):
            return self.value_ref() is other.value_ref() and self.raw_value_ref() is other.raw_value_ref()
        return False

    def __hash__(self):
        return self._hash


class DimWrapper:
    """
    Wraps :class:`Dim`.
    Using weakref for the dim.
    If the size is scalar and known, equality is given when the size is equal (and dim tag is ignored)
    """

    def __init__(self, dim: Dim, *, finalize_callback):
        self.dim_value = _dim_value_for_key(dim)
        # finalize_callback only needed when we don't use the dim value.
        self.dim_ref = ref(dim, finalize_callback if self.dim_value is None else None)
        self.dyn_size_ref = (
            # E.g. consider the batch dim or data spatial dim which would be reset each step.
            # We need some ref to the dyn size, and finalize this key when it goes out of scope.
            # This is only needed when there is no info on the static size (or eager scalar dyn size).
            ref(dim.dyn_size_ext.raw_tensor, finalize_callback)
            if self.dim_value is None and dim.dyn_size_ext is not None and dim.dyn_size_ext.raw_tensor is not None
            else None
        )
        self._hash = hash(dim) if self.dim_value is None else hash(self.dim_value)

    def __eq__(self, other):
        if isinstance(other, DimWrapper):
            if self.dim_value is not None:
                return self.dim_value == other.dim_value
            return self.dim_ref() == other.dim_ref() and self.dyn_size_ref() is other.dyn_size_ref()
        return False

    def __hash__(self):
        return self._hash


def _dim_value_for_key(dim: Dim) -> Optional[int]:
    if dim.size is not None:
        return dim.size
    if dim.dyn_size_ext is not None and not dim.dyn_size_ext.dims:
        if dim.dyn_size_ext.raw_tensor is not None:
            # noinspection PyProtectedMember
            if dim.dyn_size_ext._raw_backend.executing_eagerly():
                return int(dim.get_dim_value())
    return None


# For now... we might extend it by some more types.
_KeyItemType = Union[None, str, bool, int, float, TensorWrapper, DimWrapper]
_RawTypes = (type(None), str, bool, int, float)



"""
tests for returnn.tensor
"""

import _setup_test_env  # noqa
import sys
import unittest
from returnn.util import better_exchook
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim


def test_tensor():
    batch_dim = Dim(name="batch", dimension=None)
    time_dim = Dim(name="time", dimension=None)
    feat_dim = Dim(10)
    x = Tensor("x", (batch_dim, time_dim, feat_dim), "float32")
    print(x)


# Note: Some of the dim math tests are duplicated from test_TFUtil.py, and partly adapted.
# But here they are TF-independent.


def test_dim_math_basics():
    a = Dim(None, name="a")
    b = Dim(None, name="b")
    assert a == a
    assert (a + 2 - 2) == a
    assert a + b == a + b
    assert a + b != b + a  # not commutative
    assert a * b == a * b
    assert a * b != b * a  # not commutative
    assert 2 * a == a + a
    assert 3 * a == a + a + a
    assert a * 2 != 2 * a
    assert 2 * a + b == a + a + b
    assert a + b - b == a
    assert a + 2 * b - b + -b == a
    assert a * b + b == (a + 1) * b
    assert (a + b) * 2 == a * 2 + b * 2
    assert 0 + a + 0 == a
    assert sum([0, a, 0, a, 0]) == 2 * a


def test_dim_math_double_neg():
    a = Dim(None, name="a")
    assert --a == a


def test_dim_math_mul_rmul():
    a = Dim(None, name="a")
    b = a * 3
    c = 2 * b
    assert c == 2 * (a * 3)
    assert c != a * 6
    assert c == (2 * a) * 3


def test_dim_math_mul_div():
    a = Dim(None, name="a")
    b = Dim(None, name="b")
    assert (a * b) // b == a
    assert (b * a) // b != a
    assert (b * a).div_left(b) == a


def test_dim_math_div_mul():
    a = Dim(None, name="a")
    b = Dim(None, name="b")
    assert a // b == a // b


def test_dim_math_div_div():
    a = Dim(None, name="a")
    b = a.ceildiv_right(2)
    b = b.ceildiv_right(3)
    c = a.ceildiv_right(6)
    print(a, b, c)
    assert b == c


def test_dim_math_pad_conv():
    time = Dim(None, name="time")
    padded = 2 + time + 2
    assert padded == 2 + time + 2
    conv_valid = (-2) + padded + (-2)
    assert conv_valid == time


def test_dim_math_pad_window():
    rf.select_backend("numpy")
    batch_dim = Dim(3, name="batch")
    time1_dim = Dim(rf.convert_to_tensor([5, 6, 7], dims=[batch_dim], name="time1"))
    time2_dim = Dim(rf.convert_to_tensor([20, 21, 22], dims=[batch_dim], name="time2"))
    time3_dim = Dim(rf.convert_to_tensor([10, 5, 3], dims=[batch_dim], name="time3"))
    in_spatial_dim = time1_dim + time2_dim + time3_dim
    filter_size = 17
    # As it would happen with conv/pool/window/... with window size 17.
    out_spatial_dim = in_spatial_dim.sub_left(8).sub_right(8)
    print("out_spatial_dim:", out_spatial_dim)
    sizes = out_spatial_dim.dyn_size.tolist()
    print("sizes:", sizes)
    expected_sizes = (time1_dim.dyn_size + time2_dim.dyn_size + time3_dim.dyn_size - (filter_size - 1)).tolist()
    print("expected_sizes:", expected_sizes)
    assert sizes == expected_sizes


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
